# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-05-10 11:04


import jsonfield.fields

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('cornerstone', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='cornerstoneglobalconfiguration',
            name='subject_mapping',
            field=jsonfield.fields.JSONField(default={}, help_text='Key/value mapping cornerstone subjects to edX subjects list'),
        ),
    ]
