# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-06-13 11:07


from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0068_remove_role_based_access_control_switch'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='enterprisecustomer',
            name='catalog',
        ),
        migrations.RemoveField(
            model_name='historicalenterprisecustomer',
            name='catalog',
        ),
    ]
